/*
 * locate.c - client localization application
 *
 * Copyright (C) 2003 BLADOX, s.r.o.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <config.h>
#include <turbo/turbo.h>

#include <stdlib.h>
#include <string.h>

u8 PROGMEM t_cells[] = { "/cells" };


/* *INDENT-OFF* */

lc_char PROGMEM lc_NET[]={
        LC_EN("NET: ")
	LC_FR("RES: ")
        LC_END
};

lc_char PROGMEM lc_LACCID[]={
        LC_EN("LAC/CID: ")
        LC_END
};

/* *INDENT-ON* */

#define TAG_CELL_TEXT   	0x01
#define TAG_CELLS_COUNTRY       0x02
#define TAG_CELLS_OPERATOR      0x03
#define TAG_CELLS_DATE          0x04

#define CELLS_NO_DIR		0x00
#define CELLS_NO_LIST		0x01
#define CELLS_UNKNOWN_CELL	0x02
#define CELLS_BROKEN_FILE	0x03

u8 err_to_buf (u8 err, u8 * buf, u8 * res)
{
  u8 *r = buf;

  /*
     T_LOCATION_INFO:
     0,1,2 MCC, MNC
     3,4 LAC
     5,6 CID
   */

  r = sprints (r, locale (lc_NET));
  r = sprintch (r, swap_nibble (res[0]));
  r = sprintc (r, (swap_nibble (res[1]) >> 4) + '0');
  r = sprintc (r, '-');
  r = sprintch (r, swap_nibble (res[2]));
  r = sprintc (r, '\n');

  r = sprints (r, locale (lc_LACCID));
  r = sprintch (r, res[3]);
  r = sprintch (r, res[4]);
  r = sprintc (r, '/');
  r = sprintch (r, res[5]);
  r = sprintch (r, res[6]);
//        r = sprintc (r, '\0');
  return r - buf;
}

u8 seek_cell (u8 * res, u8 * buf1)
{
  u8 buf2[7];
  MCFILE *f = NULL;
  u16 cid, next_val, lac, tmp_lac;
  b32 b;
  b32 file_pos;
  u8 len = 0, i, j;
  u8 *operator = NULL;
  u8 *country = NULL;
  u8 *tmp;

  /*
     T_LOCATION_INFO:
     0,1,2 MCC, MNC
     3,4 LAC
     5,6 CID
   */

  buf2[0] = (res[0] & 0x0F) + '0';
  buf2[1] = ((res[0] & 0xF0) >> 4) + '0';
  buf2[2] = (res[0 + 1] & 0x0F) + '0';
  buf2[3] = (res[0 + 2] & 0x0F) + '0';
  buf2[4] = ((res[0 + 2] & 0xF0) >> 4) + '0';
  buf2[5] = '\0';

  b = cd (t_cells);
  if (b < 0)
  {
    dbsp ("no /cells\n");
    len = err_to_buf (CELLS_NO_DIR, buf1, res);
    goto X_ERR;
  }

  b = mc_find (0, buf2, FS_FIND_STRING);
  if (b < 0)
  {
    dbsp ("no cell list\n");
    len = err_to_buf (CELLS_NO_LIST, buf1, res);
    goto X_ERR;
  }

  cid = res[6] | (res[5] << 8);
  lac = res[4] | (res[3] << 8);
  dbsp ("XCID ");
  dbih (cid);
  dbc (' ');
  dbsp ("XLAC ");
  dbih (lac);
  dbc ('\n');

  f = mc_open (b, FILE_READ);

  mc_read (f, 0, buf2, 1);
  mc_read (f, 0, buf1, buf2[0] + 1);

  i = get_tag (buf1, TAG_CELLS_OPERATOR);
  if (i != 0)
  {
    operator = malloc (buf1[i + 1] + 1);

    memcpy (operator, &buf1[i + 2], buf1[i + 1]);
    operator[buf1[i + 1]] = '\0';
  }

  i = get_tag (buf1, TAG_CELLS_COUNTRY);
  if (i != 0)
  {
    country = malloc (buf1[i + 1] + 1);
    memcpy (country, &buf1[i + 2], buf1[i + 1]);
    country[buf1[i + 1]] = '\0';
  }

  file_pos = buf2[0] + 1;	//skip head

  dbsp (">seek_cid\n");

  while (1)
  {
    mc_read (f, file_pos, buf2, 3);
    next_val = buf2[0] | (buf2[1] << 8);
    len = buf2[2];

    dbsp ("cell stamp ");
    dbih (next_val);
    dbc (' ');
    dbch (len);
    dbc ('\n');

    if (next_val == 0 || cid < next_val)
      break;
    file_pos += len * 8 + 3;
  }

  file_pos += 3;

  for (i = 0; i < len; i++)
  {
    mc_read (f, file_pos + i * 8, buf2, 4);
    tmp_lac = buf2[2] | (buf2[3] << 8);
    dbsp ("LAC ");
    dbih (tmp_lac);
    dbc (' ');
    next_val = buf2[0] | (buf2[1] << 8);
    dbsp ("CID ");
    dbih (next_val);
    dbc (' ');
    dbch (i);
    dbc ('\n');

    if (cid == next_val)
    {
      next_val = 0;

      if (i + 1 < len)
      {
	mc_read (f, file_pos + (i + 1) * 8, buf2, 2);
	next_val = buf2[0] | (buf2[1] << 8);
      }

      if (cid != next_val || lac == tmp_lac)	// either both cid & lac valid or last cid (then ignore lac)
      {
	mc_read (f, file_pos + i * 8 + 4, buf2, 4);
	file_pos =
	  ((u32) buf2[0]) | ((u32) buf2[1] << 8) | ((u32) buf2[2] << 16) |
	  ((u32) buf2[3] << 24);
	dbsp ("cid found ");
	dbih (file_pos >> 16);
	dbih (file_pos);
	dbc ('\n');
	goto X_FOUND;
      }
    }
  }

  dbsp ("cid not found\n");
  len = err_to_buf (CELLS_UNKNOWN_CELL, buf1, res);
  goto X_ERR;

X_FOUND:
  dbsp ("<seek_cid\n");

  len = 0;
  tmp = buf1;

  if (operator)
  {
    i = strlen (operator);
    operator[i] = '\n';
    i++;
    memcpy (tmp, operator, i);
    tmp += i;
    len += i;
  }

  if (country)
  {
    i = strlen (country);
    country[i] = '\n';
    i++;
    memcpy (tmp, country, i);
    tmp += i;
    len += i;
  }

  /*
     tmp=sprints(tmp, t_LAC); len+=sizeof(t_LAC)-1;
     tmp=sprintch(tmp, res[3]); len+=2;
     tmp=sprintch(tmp, lac); len+=2;
     tmp=sprintc(tmp, ' '); len+=1;
   */

  mc_read (f, file_pos, tmp, 1);
  i = tmp[0];
  mc_read (f, file_pos, tmp, i + 1);

  i = get_tag (tmp, TAG_CELL_TEXT);
  if (i != 0)
  {
    u8 l1;

    dbsp ("tag found ");
    dbch (i);
    dbc ('\n');
    l1 = tmp[i + 1];
    for (j = 0; j < l1; j++)
      tmp[j] = tmp[i + j + 2];
    l1 = dcs_7b_len (l1);
    len += dcs_78 (tmp, l1, DCS_7_TO_8);
    buf1[len] = '\0';
  }
  else
    len = err_to_buf (CELLS_BROKEN_FILE, buf1, res);

X_ERR:
  mc_close (f);

  free (operator);
  free (country);
  return len;
}

static u8 PROGMEM t_TSMS_HEAD[] = "AASTAT1";

u8 rec_sms (File_apdu_data * fd)
{
  u8 *t;
  u8 *data = fd->data;
  u8 *res;
  u8 i, j, l;
  u8 *s;
  u8 buf[7];

  if ((data[0] & 0x07) == 3)	//received, to be read
  {
    s = data;
    s++;			// status
    i = *s;
    s += i + 1;			// TS-Service-Centre-Address

    i = *s;
    if ((i & 0x03) == 0x00)	//deliver
    {
      t = tpdu_seek (s, T_SMS_DCS);
      if (dcs_is_7b (*t))
      {
	t = tpdu_seek (s, T_SMS_UDL);
	i = *t;
	dcs_78 (t + 1, i, DCS_7_TO_8);
      }

      t = tpdu_seek (s, T_SMS_UDL);
      l = *t;
      t++;

      if (l > sizeof (t_TSMS_HEAD))
      {
	for (i = 0; i < sizeof (t_TSMS_HEAD) - 1; i++)
	  if (t[i] != rb (t_TSMS_HEAD + i))
	  {
	    goto X_END;
	  }
	i = sizeof (t_TSMS_HEAD) - 1;
#ifdef DEBUG
	dbsp ("DATA: ");
	for (j = 0; j < l; j++)
	{
	  dbch (t[j]);
	  dbc (' ');
	}
	dbc ('\n');
#endif
	while (t[i] != 0)
	{
	  t[i - sizeof (t_TSMS_HEAD) + 1] = t[i];
	  i++;
	}
	l = i - sizeof (t_TSMS_HEAD) + 1;
	i++;
	if (t[i] == 0)
	{
//locinfo not included, not supported
	}
	else
	{
	  dbsp ("PAD ");
	  dbch (t[i]);
	  dbc ('\n');
	  i++;			//0x7f
#ifdef DEBUG
	  dbsp ("RAW ");
	  for (j = 0; j < 14; j++)
	  {
	    dbch (t[i + j]);
	    dbc (' ');
	  }
	  dbc ('\n');
#endif
	  dbsp ("CID: ");
	  for (j = 0; j < 7; j++)
	  {
	    buf[j] = t[i];
	    i++;
	    buf[j] |= (t[i] << 4);
	    i++;
	    dbch (buf[j]);
	  }
	  dbc ('\n');
	  i = seek_cell (buf, &t[l]);
	  l = l + i;
	}
	dbsp ("LEN: ");
	dbch (l);
	dbc ('\n');
	if (l > 160)
	  l = 160;

#ifdef DEBUG
	dbsp ("NEWDATA: ");
	for (j = 0; j < l; j++)
	{
	  dbch (t[j]);
	  dbc (' ');
	}
	dbc ('\n');
#endif

      }

    X_END:
      t = tpdu_seek (s, T_SMS_DCS);
      if (dcs_is_7b (*t))
      {
	t = tpdu_seek (s, T_SMS_UDL);
	*t = l;
	dcs_78 (t + 1, l, DCS_8_TO_7);
      }
    }
  }
  store_sms_2_sim ();
}

void turbo_handler (u8 action, void *data)
{
  switch (action)
  {
    case ACTION_APP_INIT:
      reg_action (ACTION_SMS);
      break;
    case ACTION_SMS:
      stk_thread (rec_sms, data);
      break;
    default:
      break;
  }
}
